package simulator.elevatormodules;

import jSimPack.SimTime;
import simulator.framework.*;
import simulator.payloads.*;

/**
 * Models the motion of the vend carriage based on motor
 *
 * @author justinr2
 */
public class VendMotor extends Module implements TimeSensitive {

    private VendMotorPayload driveOrderedState;
    private VendCarriagePositionPayload vcPositionState;

    /**
     * Speed of the motor
     */
    public final static double MOTOR_SPEED = 200.0;  //mm/s
    public final static double MIN_POSITION = 0;
    public final static double MAX_POSITION = 700;
    private final Timer timer;
    SimTime lastRunTime;
    Direction lastDirection;

    VendMotor(SimTime period, boolean verbose) {
        super(period, "Drive", verbose);

        driveOrderedState = new VendMotorPayload();
        vcPositionState = new VendCarriagePositionPayload();
        
        driveOrderedState.set(Direction.STOP);
        vcPositionState.set(0);

        physicalConnection.sendTimeTriggered(vcPositionState, period);
        physicalConnection.registerTimeTriggered(driveOrderedState);
        physicalConnection.registerEventTriggered(driveOrderedState);

        physicalConnection.sendOnce(vcPositionState);

        //previousSpeed = driveSpeedState.speed;
        //previousDir = driveSpeedState.direction;

        lastRunTime = Harness.getTime();
        lastDirection = Direction.STOP;

        timer = new Timer(this);
        timer.start(period);
    }

    @Override
    public void receive(VendMotorPayload msg) {
        run();
    }

    public void timerExpired(Object callbackData) {
        run();
        timer.start(period);
    }

    /**
     * Calculates the current speed and position of the car.
     */
    private void run() {

        //if no time has elapsed, do not do anything
        SimTime elapsedTime = SimTime.subtract(Harness.getTime(), lastRunTime);
        if (elapsedTime.equals(SimTime.ZERO)) {
            return;
        }

        //motion dynamics - move the carriage based on how long has elapsed since
        //the last run

        double newPosition = vcPositionState.position;
        if (lastDirection != Direction.STOP) {
            //compute offset since last run based on speed and time
            double offset = this.MOTOR_SPEED * elapsedTime.getFracSeconds();
            if (lastDirection == Direction.LEFT) {
                newPosition -= offset;
            } else if (lastDirection == Direction.RIGHT) {
                newPosition += offset;
            }
            //bounds check
            newPosition = Math.max(MIN_POSITION, newPosition);
            newPosition = Math.min(MAX_POSITION, newPosition);
        }

        vcPositionState.position = newPosition;

        physicalConnection.sendOnce(vcPositionState);

        lastDirection = driveOrderedState.direction;
        lastRunTime = Harness.getTime();

        log("Last Ordered State=", driveOrderedState,
                " Car Position=", vcPositionState.position, " meters");
    }


    @Override
    public String toString() {
        return "DriveObject";
    }
}
