/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package simulator.elevatormodules.passenger;

import jSimPack.SimTime;
import java.util.ArrayList;
import java.util.Random;
import simulator.framework.Harness;
import simulator.framework.PassengerInjector;
import simulator.framework.SodaMachine;
import simulator.framework.TimeSensitive;
import simulator.framework.Timer;

/**
 *
 * @author kamikasee
 */
public class Passenger implements TimeSensitive {

    //constants
    public final static int RANDOM_ACTION_OFFSET_MIN = 1000; //ms
    public final static int RANDOM_ACTION_OFFSET_MAX = 5000; //ms
    public final static int WAIT_FOR_COIN_OFFSET_MIN = 1000; //ms
    public final static int WAIT_FOR_COIN_OFFSET_MAX = 2000; //ms
    public final static int WAIT_FOR_SODA_OFFSET_MIN = 5000; //ms
    public final static int WAIT_FOR_SODA_OFFSET_MAX = 8000; //ms
    protected final PassengerControl pc;
    protected final PassengerInjector pi;
    ArrayList<PassengerAction> actions = new ArrayList<PassengerAction>();
    PassengerAction nextAction = null;
    private int actionCount = 0;
    private final Timer timer;
    //passenger state
    private final Random rs;
    private int coinsInserted = 0;
    private int coinsReturned = 0;
    private int lastButton = SodaMachine.NONE;
    private int sodaReceived = SodaMachine.NONE;
    private long randomSeed;
    private boolean verbose;

    public Passenger(PassengerInjector pi, PassengerControl pc, long randomSeed, boolean verbose) {
        this.pi = pi;
        this.pc = pc;
        this.verbose = verbose;
        pc.setCurrentPassenger(this);
        this.randomSeed = randomSeed;
        this.rs = new Random(randomSeed);
        this.timer = new Timer(this);
    }

    public void start() {
        log("Passenger started");
        scheduleRandomAction();
    }

    void receivePassengerEvent(PassengerEvent e) {
        if (e.getType() == PassengerEventType.SODA_VENDED) {
            log("SODA_VENDING event received");
            log("no action scheduled");
            //setNextAction(new WaitForSodaAction(), WAIT_FOR_SODA_INTERVAL);
        }
    }

    //pick a random action and schedule it
    private void scheduleRandomAction() {
        SimTime randomStartOffset = computeRandomInterval(rs, RANDOM_ACTION_OFFSET_MIN, RANDOM_ACTION_OFFSET_MAX);
        PassengerAction newAction;
        switch (rs.nextInt(7)) {
            case 0:
            case 1:
            case 2:
                newAction = new PressButtonAction();
                break;
            case 3:
            case 4:
            case 5:
                newAction = new InsertCoinAction();
                break;
            case 6:
                newAction = new PressReturnAction();
                break;
            default:
                throw new RuntimeException("Inconsistent values in scheduleRandomAction switch statement.");
        }
        setNextAction(newAction, randomStartOffset);
    }

    private void setNextAction(PassengerAction action, SimTime startOffset) {
        if (timer.isRunning()) {
            //cancel pending event
            timer.cancel();
            nextAction = null;
        }
        log(String.format("Next action in %.3fs: ", startOffset.getFracSeconds()) + action.toString());
        actionCount++;
        nextAction = action;
        timer.start(startOffset);
    }

    public void timerExpired(Object callbackData) {
        PassengerAction pa = nextAction;
        actions.add(nextAction);
        nextAction.execute();
        log("Executed action: " + pa.toString());
    }

    private void log(String s) {
        if (verbose) {
            Harness.log("Passenger", s);
        }
    }

    //check the passenger state to see if we are satisfied
    public PassengerSatisfaction isSatisfied() {
        boolean isSatisfied = true;
        StringBuilder reasons = new StringBuilder();

        //check all the reasons a customer can be unsatisfied
        if (sodaReceived != SodaMachine.NONE) {
            //received a soda
            if (coinsReturned + SodaMachine.SODA_COST < coinsInserted) {
                reasons.append("  Put in " + coinsInserted + " coins but only received " + coinsReturned + " and a soda;\n");
            }
            if (lastButton != sodaReceived) {
                reasons.append("  Requested soda " + lastButton + " but recieved " + sodaReceived + "\n");
            }
        } else {
            //no soda
            if (coinsReturned < coinsInserted) {
                reasons.append("  Put in " + coinsInserted + " coins but only received " + coinsReturned + ";\n");
            }
        }
        if (reasons.length() > 0) {
            //unsatisified
            return new PassengerSatisfaction(false, reasons.toString());
        } else {
            //satisfied
            if (sodaReceived != SodaMachine.NONE) {
                return new PassengerSatisfaction(true, "  Passenger put in " + coinsInserted + " coins and received " + coinsReturned + " and soda " + sodaReceived + "\n");
            } else {
                return new PassengerSatisfaction(true, "  Passenger put in " + coinsInserted + " coins and received " + coinsReturned + " and no soda\n");
            }
        }
    }

    //call this function when the last action has been done
    private void finish() {
        //cancel any pending action
        timer.cancel();
        nextAction = null;
        PassengerSatisfaction ps = isSatisfied();
        if (ps.isSatisfied) {
            log("satisfied passenger: " + ps.reasons);
        } else {
            log("UNSATISFIED passenger: " + ps.reasons);
        }
        log("Passenger finished");
        log("############################################");
        log("############################################");
        //finished callback tells the injector we are done.
        pi.passengerFinished(this);
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        PassengerSatisfaction ps = isSatisfied();
        if (ps.isSatisfied) {
            sb.append("Customer is satisified.\n");

        } else {
            sb.append("***Customer is UNSATISFIED!");
        }
        sb.append("Reasons:\n");
        sb.append(ps.reasons);
        sb.append("Actions:  ");
        for (PassengerAction a : actions) {
            sb.append("  ");
            sb.append(a.toString());
            sb.append("\n");
        }
        return sb.toString();
    }

    /**
     * Returns a time from a uniform distribution over the given interval.
     * @param rs source of randomness
     * @param minMs minimum interval time in ms (inclusive)
     * @param maxMs maximum interval time in ms (inclusive)
     * @return a SimTime object in the specified interval.
     */
    protected final static SimTime computeRandomInterval(Random rs, int minMs, int maxMs) {
        return new SimTime(rs.nextInt(maxMs - minMs) + (minMs + 1), SimTime.SimTimeUnit.MILLISECOND);
    }

    private static enum PassengerActionType {

        PRESS_BUTTON,
        INSERT_COIN,
        PRESS_RETURN,
        WAIT_FOR_COINS,
        WAIT_FOR_SODA
    }

    private abstract class PassengerAction {

        protected SimTime executionTime = SimTime.ZERO;

        abstract PassengerActionType getType();

        void execute() {
            executionTime = Harness.getTime();
        }

        public SimTime getExecutionTime() {
            return executionTime;
        }

        @Override
        public String toString() {
            return String.format("%.3f:  ", executionTime.getFracSeconds());
        }


    }

    private class PressButtonAction extends PassengerAction {

        final static int PRESS_TIME_MIN = 150;  //in units of ms
        final static int PRESS_TIME_MAX = 300;  //in units of ms
        final SimTime pressTime;
        final int buttonIndex;

        PressButtonAction() {
            pressTime = computeRandomInterval(rs, PRESS_TIME_MIN, PRESS_TIME_MAX);
            buttonIndex = rs.nextInt(SodaMachine.NUM_CHUTES);
        }

        @Override
        PassengerActionType getType() {
            return PassengerActionType.PRESS_BUTTON;
        }

        @Override
        void execute() {
            super.execute();
            pc.sodaButtons[buttonIndex].press(pressTime);
            if (pc.buttonLights[buttonIndex].isLighted()) {
                lastButton = buttonIndex + 1;
                if (coinsInserted >= SodaMachine.SODA_COST) {
                    setNextAction(new WaitForSodaAction(), computeRandomInterval(rs, WAIT_FOR_SODA_OFFSET_MIN, WAIT_FOR_SODA_OFFSET_MAX));
                } else {
                    scheduleRandomAction();
                }
            } else {
                scheduleRandomAction();
            }
        }

        @Override
        public String toString() {
            return super.toString() + "Press " + buttonIndex + " for " + pressTime.toString() + ";";
        }
    }

    private class InsertCoinAction extends PassengerAction {

        @Override
        PassengerActionType getType() {
            return PassengerActionType.INSERT_COIN;
        }

        @Override
        void execute() {
            super.execute();
            pc.changeMachine.insertCoin();
            coinsInserted++;
            scheduleRandomAction();
        }

        @Override
        public String toString() {
            return super.toString() + "Insert coin;";
        }
    }

    private class PressReturnAction extends PassengerAction {

        final static int PRESS_TIME_MIN = 150;  //in units of ms
        final static int PRESS_TIME_MAX = 300;  //in units of ms
        final SimTime pressTime;

        PressReturnAction() {
            pressTime = computeRandomInterval(rs, PRESS_TIME_MIN, PRESS_TIME_MAX);
        }

        @Override
        PassengerActionType getType() {
            return PassengerActionType.PRESS_RETURN;
        }

        @Override
        void execute() {
            super.execute();
            pc.coinReturnButton.press(pressTime);
            setNextAction(new WaitForCoinAction(), computeRandomInterval(rs, WAIT_FOR_COIN_OFFSET_MIN, WAIT_FOR_COIN_OFFSET_MAX));
        }

        @Override
        public String toString() {
            return super.toString() + "Press CoinReturn for " + pressTime.toString() + ";";
        }
    }

    private class WaitForCoinAction extends PassengerAction {

        @Override
        PassengerActionType getType() {
            return PassengerActionType.WAIT_FOR_COINS;
        }

        @Override
        void execute() {
            super.execute();
            coinsReturned += pc.changeMachine.getCoinsFromTray();
            finish();
        }

        @Override
        public String toString() {
            return super.toString() + "Wait for coins;";
        }
    }

    private class WaitForSodaAction extends PassengerAction {

        @Override
        PassengerActionType getType() {
            return PassengerActionType.WAIT_FOR_SODA;
        }

        @Override
        void execute() {
            super.execute();
            coinsReturned += pc.changeMachine.getCoinsFromTray();  //get any extra change from the tray
            sodaReceived = pc.vendActuator.getSodaFromTray();
            finish();
        }

        @Override
        public String toString() {
            return super.toString() + "Wait for soda;";
        }
    }

    public static class PassengerSatisfaction {

        public final boolean isSatisfied;
        public final String reasons;

        private PassengerSatisfaction(boolean isSatisfied, String reasons) {
            this.isSatisfied = isSatisfied;
            this.reasons = reasons;
        }
    }

    public static void main(String args[]) {
        Random rs = new Random(4242);
        for (int i=0; i < 100; i++) {
            System.out.println(Passenger.computeRandomInterval(rs, RANDOM_ACTION_OFFSET_MIN, RANDOM_ACTION_OFFSET_MAX));
        }
    }
}
