package simulator.framework;

import jSimPack.*;
import java.util.*;
import java.io.*;

/**
 * Allows the elevator components to interact with each other through a shared
 * event queue.
 *
 * @author Christopher Martin
 * @author Kenny Stauffer
 */
public class Harness { //implements FutureEventListener {

    private static final boolean verbose = false;
    private static final RandomSource randomSource;
    private static PhysicalNetwork thePhysicalNetwork;
    private static CANNetwork theCANNetwork;
    /**
     * The HarnessEventListener objects that have registered to receive events
     * of a particular type.
     */
    private static final Map<Integer, Collection<HarnessEventListener>> listeners;
    private static final FutureEventList eventList;
    private static PrintStream logPrinter = System.out;
    private static final Harness theHarness;

    static {
        theHarness = new Harness();
        randomSource = new RandomSource();
        eventList = new FutureEventList(randomSource);
        listeners = new HashMap<Integer, Collection<HarnessEventListener>>();
    }

    public static void initialize(SimTime canBitTime,
            boolean networkVerbose, boolean frameworkNetworkVerbose,
            boolean utilizationVerbose, boolean dropVerbose) {
        thePhysicalNetwork = new PhysicalNetwork();
        theCANNetwork = new CANNetwork(canBitTime);
        theCANNetwork.getUtilization().setVerbose(utilizationVerbose);
        theCANNetwork.setVerbose(networkVerbose);
        theCANNetwork.setDropVerbose(dropVerbose);
        thePhysicalNetwork.setVerbose(frameworkNetworkVerbose);
        //thePhysicalNetwork.setDropVerbose(dropVerbose);
    }

    //interface to add and remove network events
    public static FutureEvent schedule(FutureEventListener event, SimTime timeInterval, Object data) {
        return eventList.schedule(event, timeInterval, data);
    }

    public static FutureEvent scheduleNonsimulationEvent(FutureEventListener event, SimTime timeInterval, Object data) {
        return eventList.scheduleNonsimulationEvent(event, timeInterval, data);
    }

    public static boolean cancelNonsimulationEvent(FutureEvent event, boolean deleteNow) {
        return eventList.cancelNonsimulationEvent(event, deleteNow);
    }

    public static boolean cancelEvent(FutureEvent event, boolean deleteNow) {
        return eventList.cancelEvent(event, deleteNow);
    }

    /**
     * Directs the harness to broadcast events of the specified type to the
     * specified <code>HarnessEventListener</code>.
     */
    /*public static void addListener(HarnessEventListener who, int eventType) {
    if (verbose) {
    log("Harness","addListener(",who,",",eventType,")");
    }
    Collection<HarnessEventListener> thisListener = listeners.get(eventType);
    if (thisListener == null) {
    thisListener = new LinkedHashSet<HarnessEventListener>();
    listeners.put(eventType, thisListener);
    }
    thisListener.add(who);
    }*/

    /*public static void enqueue(HarnessEvent e) {
    if (verbose) {
    log("Harness","enqueue(",e,")");
    }
    e.validate();
    SimTime scheduleTime = e.getDispatchDelay();
    eventList.schedule(theHarness, scheduleTime, e);
    }*/
    /**
     * Removes the specified event from the event list.  If the event has not
     * happened yet, it will not happen in the future, even if it was about to
     * happen during the next simulation step.
     * 
     * @return
     * <code>true</code> if and only if the event was removed from the event list
     */
    /*public static boolean cancel(HarnessEvent e) {
    if (!e.isValid())
    return false;
    e.invalidate();
    return true;
    }*/

    /*public void eventReleased(Object callback) {
    if (verbose) {
    log("Harness","enter eventReleased(",callback,")");
    }
    HarnessEvent e = (HarnessEvent) callback;
    if (!e.isValid())
    return;
    if (e.getSource() instanceof Timer) {
    // this is an awful hack to avoid traversing a linked list of
    // Timers that are all registered for the same event type.
    e.getSource().eventDispatched(e);
    return;
    }
    Collection<HarnessEventListener> thisListeners = listeners.get(e.getType());
    if (thisListeners != null) {
    for (HarnessEventListener listener : thisListeners) {
    if (verbose) {
    log("Harness", "calling ",listener,".eventDispatched(",e,")");
    }
    listener.eventDispatched(e);
    }
    }
    }*/
    public static void runSim(SimTime howLong) {
        eventList.runSimulationUntil(howLong);
    }

    public static void endSim() {
        eventList.endSimulation();
    }

    public static void endSimAt(SimTime endTime) {
        eventList.setEndTime(endTime);
    }

    public static void setRealtimeRate(double rate) {
        eventList.setRealtimeRate(rate);
    }

    public static double getRealtimeRate() {
        return eventList.getRealtimeRate();
    }

    public static void stepSimulation() {
        eventList.stepSimulation();
    }

    public static boolean simulationIsBlocked() {
        return eventList.isBlocked();
    }

    /**
     * Returns how long the simulation has been running, in simulated time.
     */
    public static SimTime getTime() {
        return eventList.getWallClock();
    }

    public static RandomSource getRandomSource() {
        return randomSource;
    }

    public static void setRandomSeed(long seed) {
        randomSource.setSeed(seed);
    }

    public static long getRandomSeed() {
        return randomSource.getSeed();
    }

    public static void setLogFilename(String filename)
            throws FileNotFoundException {
        logPrinter = new PrintStream(filename);
    }

    /**
     * Prints a log message that includes the source and timestamp of the
     * message.
     * 
     * @param source
     * The object that generated this log message.
     * 
     * @param msg
     * The log message to print.
     */
    public static void log(String source, Object... msg) {
        logPrinter.format("[%s] @%4.9f: ", source, getTime().getFracSeconds());
        for (Object o : msg) {
            logPrinter.print(o);
        }
        logPrinter.println();
    }

    //breakpoint methods
    /**
     * Add a breakpoint at the specified time.  Specifying a time in the past will
     * cause an exception.  The breakpoint stops execution before any events that
     * occur at that time.
     * 
     * @param breakpointTime  The time of the breakpoint.
     * @return true if the breakpoint was added, false if there was already a 
     * breakpoint at that time.
     */
    public static boolean addBreakpoint(SimTime breakpointTime) {
        return eventList.addBreakpoint(breakpointTime);
    }

    public static boolean removeBreakpoint(SimTime breakpointTime) {
        return eventList.removeBreakpoint(breakpointTime);
    }

    public static void addBreakpointListener(BreakpointListener l) {
        eventList.addBreakpointListener(l);
    }

    public static boolean removeBreakpointListener(BreakpointListener l) {
        return eventList.removeBreakpointListener(l);
    }

    static void interleaveLock() {
        eventList.interleaveLock();
    }

    static void interleaveUnlock() {
        eventList.interleaveUnlock();
    }
    
    
    
    public static PhysicalNetwork getPhysicalNetwork() {
        return thePhysicalNetwork;
    }

    public static CANNetwork getCANNetwork() {
        return theCANNetwork;
    }

    private Harness() {
    }
}
