package simulator.framework;

import jSimPack.SimTime;
//import simulator.payloads.translators.CanPayloadTranslator;
import java.lang.reflect.*;
import java.util.*;
import java.text.ParseException;

import java.util.Map.Entry;
import simulator.payloads.*;

public class MessageInjector extends Networkable implements TimeSensitive, Parser {

    public final static String DEFINE_DIRECTIVE = "#DEFINE";
    private ArrayList<String> assertionResults = new ArrayList<String>();
    private HashMap<String, String> macros = new HashMap<String, String>();
    private int assertionPassCount = 0;
    private int assertionFailCount = 0;
    private int assertionTotal = 0; // this is incremented when the assertions are created, not when they are executed.

    public String getAssertionStats() {
        StringBuffer sb = new StringBuffer("Assertion Results:\n");
        for (String s : assertionResults) {
            sb.append(s);
            sb.append("\n");
        }
        sb.append("\n\n");
        sb.append(getAssertionSummary());
        return sb.toString();
    }

    public String getAssertionSummary() {
        StringBuffer sb = new StringBuffer("******************Summary******************\n");
        sb.append("Passed:  " + assertionPassCount + "\n");
        sb.append("Failed:  " + assertionFailCount + "\n");
        sb.append("Total :  " + assertionTotal + "\n");

        if (assertionTotal != assertionPassCount + assertionFailCount) {
            sb.append("\n  Some assertions were not executed, so you should probably increase the runtime.  \n");
        }
        return sb.toString();
    }

    private interface Executable {

        public void execute();
    }

    private enum Operator {

        EQUAL("==", false),
        NOT_EQUAL("!=", false),
        LESS_THAN("<", true),
        GREATER_THAN(">", true),
        LESS_THAN_OR_EQUAL("<=", true),
        GREATER_THAN_OR_EQUAL(">=", true);
        private String stringRep;
        private boolean requiresNumericTypes;

        private Operator(String stringRep, boolean requiresNumericTypes) {
            this.stringRep = stringRep;
            this.requiresNumericTypes = requiresNumericTypes;
        }

        public boolean getRequiresNumericTypes() {
            return requiresNumericTypes;
        }

        public String getStringRepresentation() {
            return stringRep;
        }

        @Override
        public String toString() {
            return stringRep;
        }
    }

    private class InjectablePayload implements Executable {

        /**
         * The message to inject.
         */
        final public Payload msg;
        /**
         * The interval between broadcasts.
         */
        final public SimTime period;
        /**
         * The network on which to broadcast this message.
         */
        final public MessageContext messageContext;

        public InjectablePayload(Payload msg, SimTime period, MessageContext messageContext) {
            this.msg = msg;
            this.period = period;
            this.messageContext = messageContext;
        }

        @Override
        public String toString() {
            return "InjectablePayload:" + msg;
        }

        public void execute() {
            logPrint("injecting " + msg + " with period " + period);

            switch (messageContext) {
                case PHYSICAL:
                    if (period.isPositive()) {
                        logPrint("sending on FI");
                        fi.sendTimeTriggered(msg, period);
                    } else {
                        logPrint("sending single event on FI");
                        fi.sendOnce(msg);
                    }
                    break;
                case NETWORK:
                    if (period.isPositive()) {
                        logPrint("sending on CNI");
                        cni.sendTimeTriggered(msg, period);
                    } else {
                        logPrint("sending single event on CNI");
                        cni.sendOnce(msg);
                    }
                    break;
                default:
                    throw new RuntimeException("injected message context " + messageContext + " should be NETWORK or FRAMEWORK.");
            }
        }
    }

    private class AssertionMonitor implements Executable {

        final public Object target; //this is a translator or payload
        final public Method method;
        final public Operator operator;
        final public Field field;
        final public Object value;

        public AssertionMonitor(Object target,
                Method method,
                Operator operator,
                Object value) {
            this.target = target;
            this.method = method;
            this.field = null;
            this.operator = operator;
            this.value = value;
            if (method == null) {
                throw new NullPointerException("Method");
            }
            assertionTotal++;
        }

        public AssertionMonitor(Object target,
                Field field,
                Operator operator,
                Object value) {
            this.target = target;
            this.method = null;
            this.field = field;
            this.operator = operator;
            this.value = value;
            if (field == null) {
                throw new NullPointerException("Field");
            }
            assertionTotal++;
        }

        public void execute() {
            logPrint("Testing assertion " + target + " = " + value);
            if (method != null) {
                try {
                    testValues(method.invoke(target, (Object[]) null), value);
                } catch (IllegalAccessException ex) {
                    throw new RuntimeException("Could not read method " + method + " on " + target + ": " + ex);
                } catch (IllegalArgumentException ex) {
                    throw new RuntimeException("Could not read method " + method + " on " + target + ": " + ex);
                } catch (InvocationTargetException ex) {
                    throw new RuntimeException("Could not read method " + method + " on " + target + ": " + ex);
                }
            }
            if (field != null) {
                try {
                    testValues(field.get(target), value);
                } catch (IllegalAccessException ex) {
                    throw new RuntimeException("Could not read field " + field + " on " + target + ": " + ex);
                } catch (IllegalArgumentException ex) {
                    throw new RuntimeException("Could not read field " + field + " on " + target + ": " + ex);
                }
            }
        }

        /**
         * perform a operator b
         * @param a
         * @param b
         */
        private void testValues(Object a, Object b) {
            switch (operator) {
                case EQUAL:
                    if (a.equals(b)) {
                        pass();
                    } else {
                        fail();
                    }
                    return;
                case NOT_EQUAL:
                    if (!a.equals(b)) {
                        pass();
                    } else {
                        fail();
                    }
                    return;
            }
            //if we get this far, we have a numeric type
            int comparison = 0;
            if ((a instanceof Double) && (b instanceof Double)) {
                comparison = ((Double) a).compareTo((Double) b);
            } else if ((a instanceof Integer) && (b instanceof Integer)) {
                comparison = ((Integer) a).compareTo((Integer) b);
            } else if ((a instanceof Byte) && (b instanceof Byte)) {
                comparison = ((Byte) a).compareTo((Byte) b);
            } else if ((a instanceof Float) && (b instanceof Float)) {
                comparison = ((Float) a).compareTo((Float) b);
            } else if ((a instanceof Long) && (b instanceof Long)) {
                comparison = ((Long) a).compareTo((Long) b);
            } else if ((a instanceof Short) && (b instanceof Short)) {
                comparison = ((Short) a).compareTo((Short) b);
            } else {
                throw new RuntimeException("Could not convert values " + a + " and " + b + " for operator " + operator.getStringRepresentation());
            }
            switch (operator) {
                case GREATER_THAN:
                    if (comparison > 0) {
                        pass();
                    } else {
                        fail();
                    }
                    break;
                case GREATER_THAN_OR_EQUAL:
                    if (comparison >= 0) {
                        pass();
                    } else {
                        fail();
                    }
                    break;
                case LESS_THAN:
                    if (comparison < 0) {
                        pass();
                    } else {
                        fail();
                    }
                    break;
                case LESS_THAN_OR_EQUAL:
                    if (comparison <= 0) {
                        pass();
                    } else {
                        fail();
                    }
                    break;
                default:
                    break;
            }
        }

        private void pass() {
            String s = "PASSED " + target + " :: " + getMemberName() + " ?" + operator.getStringRepresentation() + " " + value + " PASSED";
            String m = fillString('*', s.length());
            //String marker = new StringBuffer()
            assertionResults.add("@" + Harness.getTime() + ":  " + s);
            assertionPassCount++;
            Harness.log("AssertionMonitor", m);
            Harness.log("AssertionMonitor", s);
            Harness.log("AssertionMonitor", m);
        }

        private String fillString(char c, int length) {
            StringBuffer b = new StringBuffer(length);
            for (int i=0; i < length; i++) {
                b.append(c);
            }
            return b.toString();
        }
        
        private void fail() {
            String s = "FAILED " + target + " :: " + getMemberName() + " ?" + operator.getStringRepresentation() + " " + value + " FAILED";
            String m1 = fillString('/', s.length());
            String m2 = fillString('\\', s.length());
            assertionFailCount++;
            assertionResults.add("@" + Harness.getTime() + ":  " + s);
            Harness.log("AssertionMonitor", m1);
            Harness.log("AssertionMonitor", s);
            Harness.log("AssertionMonitor", m2);
        //throw new RuntimeException("Assertion " + target + " = " + value + " FAILED");
        }
        
        private String getMemberName() {
            if (method != null) {
                return method.getName();
            } else {
                return field.getName();
            }
        }
    }
    boolean verbose;
    ReflectionFactory refFactory;
    FileTokenizer fp;
    SimTime lastInjectionTime;
    NetworkConnection cni;
    NetworkConnection fi;

    public MessageInjector(String filename, boolean verbose) {
        this.verbose = verbose;
        lastInjectionTime = SimTime.ZERO;
        cni = Harness.getCANNetwork().getFrameworkConnection(this);
        fi = Harness.getPhysicalNetwork().getFrameworkConnection(this);

        refFactory = new ReflectionFactory(verbose);

        fp = new FileTokenizer(filename, verbose, this);
        fp.parseFile();
    }

    public SimTime lastInjectionTime() {
        return lastInjectionTime;
    }

    private class ParseTracker {

        public String[] words;
        public int currentWord = 0;

        public ParseTracker(String[] words) {
            this.words = words;
        }

        public String getNextWord() throws ParseException {
            if (!hasNextWord()) {
                throw new ParseException("Not enough arguments in " + words, currentWord);
            }
            String retval = words[currentWord];
            currentWord++;
            return retval;
        }

        public String getPreviousWord() {
            return words[currentWord - 1];
        }

        public String peekNextWord() {
            return words[currentWord];
        }

        public void increment() {
            currentWord++;
        }

        public boolean hasNextWord() {
            return (currentWord < words.length);
        }

        /**
         * Incremetns currentWord pointer if compare successful, otherwise just 
         * return false.
         */
        public boolean compareNextWord(String compareTo) {
            if (words[currentWord].equals(compareTo)) {
                currentWord++;
                return true;
            } else {
                return false;
            }
        }
    }

    /**
     * Constructs a message from the specified parameters, and queues the message
     * to be sent at a later time.  The format for the textual description of
     * an injected message is explained in the usage message of Elevator.main.
     */
    public void parse(String[] words) throws ParseException {
        logPrint(fp.lineMessage("parsing " + Arrays.toString(words)));

        if (words[0].equals(DEFINE_DIRECTIVE)) {
            if (words.length != 3) {
                throw new ParseException("Expected two arguments for a #DEFINE macro.", 0);
            }
            //add the macro to the list
            macros.put(words[1], words[2]);
            return;
        }
        
        if (words.length < 4) {
            throw new ParseException(words.length +
                    " is not enough parameters.  Need at least four (time, period, N or F, type).", 0);
        }

        //do macro replacements
        if (macros.size() > 0) {
            for (int i=0; i < words.length; i++) {
                for (Entry<String, String> macro : macros.entrySet()) {
                    if (words[i].equals(macro.getKey())) {
                        words[i] = macro.getValue();
                        break;
                    }
                }
            }
        }

        ParseTracker pt = new ParseTracker(words);

        SimTime injectionTime = new SimTime(pt.getNextWord());

        if (injectionTime.isNegative()) {
            throw new ParseException("injection time must not be negative.", 0);
        }

        if (injectionTime.isGreaterThan(lastInjectionTime)) {
            lastInjectionTime = injectionTime;
        }

        logPrint(fp.lineMessage("injection time: " + injectionTime));

        if (pt.compareNextWord("I")) {
            createInjectablePayload(pt, injectionTime);
        } else if (pt.compareNextWord("A")) {
            createAssertionMonitor(pt, injectionTime);
        } else {
            throw new ParseException(
                    "expected type \"A\" or \"I\", received " + pt.getNextWord(), 0);
        }
    }

    private void createInjectablePayload(ParseTracker pt, SimTime injectionTime) throws ParseException {
        SimTime period = new SimTime(pt.getNextWord());

        logPrint(fp.lineMessage("period: " + period));

        if (period.isNegative()) {
            throw new ParseException("period must be non-negative.", 0);
        }

        MessageContext messageContext;
        List<String> constructorArgs = new ArrayList<String>();
        List<String> packagePath = new ArrayList<String>();
        Payload payload = null;

        if (pt.compareNextWord("F")) {
            messageContext = MessageContext.PHYSICAL;
            constructorArgs.add(pt.getNextWord() + "Payload");
            packagePath.add("simulator.payloads.");
        } else if (pt.compareNextWord("N")) {
            messageContext = MessageContext.NETWORK;
            int type;
            try {
                if (pt.peekNextWord().startsWith("0x")) {
                    type = Integer.parseInt(pt.getNextWord().substring(2), 16);
                } else {
                    type = Integer.parseInt(pt.getNextWord());
                }
            } catch (NumberFormatException e) {
                throw new ParseException(
                        "expected integer for message type, received " + pt.getPreviousWord(), pt.currentWord - 1);
            }
            payload = new CanMailbox(type);
            constructorArgs.add(pt.getNextWord() + "CanPayloadTranslator");
            // A Translator takes a CanMailbox object as its first parameter.
            // We add null to parameter list so the reflection factory
            // can find the constructor.
            constructorArgs.add(null);
            packagePath.add("simulator.elevatorcontrol.");
            packagePath.add("simulator.elevatormodules.");
            packagePath.add("simulator.payloads.translators.");
        } else {
            throw new ParseException(
                    "expected network type \"F\" or \"N\", received " + pt.getNextWord(), pt.currentWord - 1);
        }

        /* next field should be the class name, so start at the next field to
         * find the "="
         */
        while (pt.hasNextWord()) {
            if (pt.compareNextWord("=")) {
                break;
            }
            constructorArgs.add(pt.getNextWord());
        }
        //put the rest of the arguments into the setArgs list
        List<String> setArgs = new ArrayList<String>();
        while (pt.hasNextWord()) {
            setArgs.add(pt.getNextWord());
        }

        logPrint(fp.lineMessage("Constructor args: " + constructorArgs));
        logPrint(fp.lineMessage("set() args: " + setArgs));

        List<Object> parameters = new ArrayList<Object>();
        Constructor<?> con = null;
        con = refFactory.prepareConstructor(packagePath,
                constructorArgs, parameters);

        //constructe t
        Object translator = null;
        try {
            if (payload == null) {
                //create payload for framework payload object
                payload = (Payload) con.newInstance(parameters.toArray());
            } else {
                //create a translator object using the payload previously created
                parameters.set(0, payload);
                translator = con.newInstance(parameters.toArray());
            }
        } catch (Exception e) {
            throw new RuntimeException("could not call constructor", e);
        }

        try {
            Method setMethod = refFactory.getMethod(con.getDeclaringClass(),
                    "set", setArgs);
            refFactory.invoke(setMethod,
                    translator == null ? payload : translator,
                    setArgs);
        } catch (NoSuchMethodException e) {
            throw new ParseException("could not find " +
                    con.getDeclaringClass().getName() +
                    ".set() that accepts: " + setArgs, 0);
        } catch (SecurityException e) {
            throw new RuntimeException(
                    constructorArgs.get(0) + ".set() could not be invoked",
                    e);
        } catch (Exception e) {
            throw new RuntimeException(constructorArgs.get(0) +
                    ".set() failed.", e);
        }

        InjectablePayload imp =
                new InjectablePayload(payload, period, messageContext);

        new SystemTimer(this).start(injectionTime, imp);

    }

    private void createAssertionMonitor(ParseTracker pt, SimTime injectionTime) throws ParseException {

        //no period for assertion monitor

        //MessageContext messageContext;
        List<String> constructorArgs = new ArrayList<String>();
        List<String> packagePath = new ArrayList<String>();
        Payload payload = null;

        if (pt.compareNextWord("F")) {
            //messageContext = MessageContext.PHYSICAL;
            constructorArgs.add(pt.getNextWord() + "Payload");
            packagePath.add("simulator.payloads.");
        } else if (pt.compareNextWord("N")) {
            //messageContext = MessageContext.NETWORK;
            int type;
            try {
                if (pt.peekNextWord().startsWith("0x")) {
                    type = Integer.parseInt(pt.getNextWord().substring(2), 16);
                } else {
                    type = Integer.parseInt(pt.getNextWord());
                }
            } catch (NumberFormatException e) {
                throw new ParseException(
                        "expected integer for message type, received " + pt.getPreviousWord(), pt.currentWord - 1);
            }
            payload = new CanMailbox(type);
            cni.registerTimeTriggered(payload);
            constructorArgs.add(pt.getNextWord() + "CanPayloadTranslator");
            // A Translator takes a CanMailbox object as its first parameter.
            // We add null to parameter list so the reflection factory
            // can find the constructor.
            constructorArgs.add(null);
            packagePath.add("simulator.elevatorcontrol.");
            packagePath.add("simulator.elevatormodules.");
            packagePath.add("simulator.payloads.translators.");
        } else {
            throw new ParseException(
                    "expected network type \"F\" or \"N\", received " + pt.getNextWord(), pt.currentWord - 1);
        }

        /* get the rest of the constructor arguments up to the : */
        while (pt.hasNextWord()) {
            if (pt.compareNextWord(":")) {
                break;
            }
            constructorArgs.add(pt.getNextWord());
        }

        //the remaining arguments are the name of the member to check and the value to check

        String memberStr = pt.getNextWord();
        String opStr = pt.getNextWord();
        Operator operator = null;
        for (Operator o : Operator.values()) {
            if (o.getStringRepresentation().equals(opStr)) {
                operator = o;
                break;
            }
        }
        if (operator == null) {
            throw new ParseException("Expected an operator, not \"" + opStr + "\"", (pt.currentWord - 1));
        }
        String valueStr = pt.getNextWord();

        logPrint(fp.lineMessage("Constructor args: " + constructorArgs));
        logPrint(fp.lineMessage("Member to Test: " + memberStr));
        logPrint(fp.lineMessage("Value to Test: " + valueStr));

        List<Object> parameters = new ArrayList<Object>();
        Constructor<?> con = null;
        con = refFactory.prepareConstructor(packagePath,
                constructorArgs, parameters);

        //constructe t
        Object translator = null;
        try {
            if (payload == null) {
                //create payload for framework payload object
                payload = (Payload) con.newInstance(parameters.toArray());
                fi.registerTimeTriggered(payload);
            } else {
                //create a translator object using the payload previously created
                parameters.set(0, payload);
                translator = con.newInstance(parameters.toArray());
            }
        } catch (Exception e) {
            throw new RuntimeException("could not call constructor", e);
        }

        /* now we have a translator or payload and we need to identify the member
         * either a field or method with no args.
         */

        Object target = null;
        if (translator == null) {
            target = payload;
        } else {
            target = translator;
        }

        AssertionMonitor mon = null;
        try {
            Method testMethod = refFactory.getMethod(con.getDeclaringClass(), memberStr);
            Object testValue = refFactory.createReturnType(testMethod, valueStr);
            testNumericType(testMethod.getReturnType(), operator);
            mon = new AssertionMonitor(target, testMethod, operator, testValue);
        } catch (NoSuchMethodException e) {
            logPrint(fp.lineMessage("could not find method " +
                    con.getDeclaringClass().getName() +
                    "." + memberStr + "()."));
        }

        if (mon == null) {
            //no method, so try a field
            try {
                Field testField = refFactory.getField(con.getDeclaringClass(), memberStr);
                Object testValue = refFactory.createFieldType(testField, valueStr);
                testNumericType(testField.getType(), operator);
                mon = new AssertionMonitor(target, testField, operator, testValue);
            } catch (NoSuchFieldException ex) {
                logPrint(fp.lineMessage("could not find field " +
                        con.getDeclaringClass().getName() +
                        "." + memberStr + "."));
            }
        }

        if (mon == null) {
            throw new ParseException("Could not find a method or field in " +
                    con.getDeclaringClass().getName() +
                    " named " + memberStr + ".", 0);
        }

        new SystemTimer(this).start(injectionTime, mon);
    }

    private void testNumericType(Type t, Operator o) throws ParseException {
        if (o.getRequiresNumericTypes()) {
            if (t != Double.TYPE &&
                    t != Float.TYPE &&
                    t != Integer.TYPE &&
                    t != Long.TYPE &&
                    t != Short.TYPE &&
                    t != Byte.TYPE) {
                throw new ParseException("Operator " + o.getStringRepresentation() + " requires numeric type, not " + t + ".", 0);
            }
        }
    }

    public void timerExpired(Object msg) {
        logPrint("enter timerExpired(" + msg + ")");

        ((Executable) msg).execute();

    }

    private void logPrint(String s) {
        if (verbose) {
            Harness.log("MessageInjector", s);
        }
    }

    @Override
    public String toString() {
        return "MessageInjector";
    }
}
