package simulator.framework;

import jSimPack.SimTime;

import java.util.*;
import java.lang.reflect.*;

/**
 * A facade that provides a simple interface for instantiating objects, finding
 * methods, and invoking methods.
 * 
 * @author Kenny Stauffer
 */
public class ReflectionFactory
{
    boolean verbose;

    public ReflectionFactory()
    {
        this(false);
    }

    public ReflectionFactory(boolean verbose)
    {
        this.verbose = verbose;
    }

    public Object instantiate(List<String> words)
    {
        List<String> path = Collections.emptyList();
        List<Object> params = new ArrayList<Object>();
        Constructor<?> con = prepareConstructor(path, words, params);
        try
        {
            return con.newInstance(params.toArray());
        }
        catch (Exception e)
        {
            System.err.println(e);
            e.printStackTrace();
        }
        System.exit(1);
        return null;
    }

    /**
     * Attempts to instantiate an object from strings representing its type and
     * parameter values. The first FutureEvent in <code>words</code> is the
     * fully qualified name of the class of the desired object, and the rest of
     * the elements are the parameter values to pass to the constructor. If the
     * specified class has a public constructor that accepts the same number of
     * specified, and each specified parameter can be converted to the
     * corresponding formal parameter in the constructor, this method will
     * construct and return the object.
     * 
     * @param packagePath
     *        packages in which to look for the named class
     * @param words
     *        The fully qualified name of the class of the object to construct,
     *        followed by the values to pass to the appropriate constructor.
     * @param parameters
     *        As the parameters are converted to objects, they will be appened
     *        to this list.
     * @return A <code>Constructor</code> that can be invoked with the
     *         parameters in <code>parameters</code>. values.
     */
    public Constructor<?> prepareConstructor(List<String> packagePath,
            List<String> words, List<Object> parameters)
    {
        log("instantiate(" + packagePath + "," + words + ")");

        if (words == null)
            throw new NullPointerException("words");
        if (words.size() == 0)
            throw new IllegalArgumentException(
                    "expected a class name as first word.");
        
        Class<?> c = null;
        if (packagePath == null || packagePath.size() == 0)
            try
            {
                c = Class.forName(words.get(0));
            }
            catch (ClassNotFoundException e)
            {
            }
        if(packagePath!=null) for (String pack : packagePath)
        {
            try
            {
                //System.out.println(pack + words.get(0));
                c = Class.forName(pack + words.get(0));
            }
            catch (ClassNotFoundException e)
            {
            }
        }
        if (c == null)
            throw new IllegalArgumentException("could not find class \""
                    + words.get(0) + "\"");

        log("found a class: " + c.getName());

        Constructor<?>[] cons = c.getConstructors();
        if (cons.length == 0)
        {
            throw new IllegalArgumentException(words.get(0)
                    + " has no public constructor");
        }

        Constructor<?> theConstructor = null;
        Class<?>[] paramTypes = null;
        for (Constructor<?> con : cons)
        {
            paramTypes = con.getParameterTypes();
            log("found constructor with " + paramTypes.length + " parameters.");

            if (paramTypes.length != words.size() - 1)
            {
                continue;
            }

            // Convert the string representation of each argument into an
            // Object to pass to the constructor.
            boolean match = true;
            for (int i = 0; i < paramTypes.length; ++i)
            {
                try
                {
                    parameters.add(parseString(paramTypes[i], words.get(i
                            + 1)));
                }
                catch (Exception e)
                {
                    log("could not convert argument "+i+" from "+words.get(i+1)
                            + " to "+paramTypes[i]+": "+e);
                    match = false;
                    parameters.clear();
                    break;
                }
            }
            if (match)
            {
                theConstructor = con;
                break;
            }
        }

        if (null == theConstructor)
        {
            throw new IllegalArgumentException("didn't find "
                    + "constructor that accepts this number of arguments.");
        }

        log("These words: " + words);
        log("     Were parsed as these types: " + Arrays.toString(paramTypes));
        log("     Which produced these objects: " + parameters);
        log("     Which will be given to this constructor: "
                + theConstructor.toGenericString());

        return theConstructor;
    }

    
    /**
     * Use this getMethod call for methods with no arguments
     * @param theClass
     * @param methodName
     * @return
     * @throws java.lang.NoSuchMethodException
     */
    public Method getMethod(Class<?> theClass, String methodName)
            throws NoSuchMethodException
    {
        log("getMethod("+theClass+", "+methodName+")");
        for (Method m : theClass.getMethods())
        {
            if (!m.getName().equals(methodName))
                continue;
            Class<?>[] argTypes = m.getParameterTypes();
            if (argTypes.length != 0)
                //throw new IllegalArgumentException("Method " + methodName + " exists but has " + argTypes.length + " arguments (expected none)");
                //can't error on the argument length, because there might be two methods with the same name.
                continue;
            return m;
        }
        throw new NoSuchMethodException("No method named " + methodName + " with 0 arguments found.");
    }

    
    public Method getMethod(Class<?> theClass, String methodName, List<String> args)
            throws NoSuchMethodException
    {
        log("getMethod("+theClass+", "+methodName+", "+args+")");
        for (Method m : theClass.getMethods())
        {
            if (!m.getName().equals(methodName))
                continue;
            Class<?>[] argTypes = m.getParameterTypes();
            if (argTypes.length != args.size())
                continue;
            for (int i = 0; i < argTypes.length; ++i)
            {
                try
                {
                    parseString(argTypes[i], args.get(i));
                }
                catch (Exception e)
                {
                    throw new IllegalArgumentException(m + " expects a "
                            + argTypes[i].getName() + " for parameter " + i
                            + ", but received " + args.get(i));
                }
            }
            return m;
        }
        throw new NoSuchMethodException();
    }

    public Field getField(Class<?> theClass, String fieldName) throws NoSuchFieldException {
        return theClass.getDeclaredField(fieldName);
    }

    public Object getStaticFieldValue(Class<?> sourceClass, String fieldName) throws NoSuchFieldException, IllegalArgumentException, IllegalAccessException {
        Field idField = getField(sourceClass, fieldName);
        return idField.get(null);
    }
    
    /**
     * Tries to create an object of the type returned by the method using the 
     * given value string.
     */
    public Object createReturnType(Method theMethod, String value) {
        return parseString(theMethod.getReturnType(), value);
    }
    
    public Object createFieldType(Field field, String value) {
        return parseString(field.getType(), value);
    }
    
    
    public Object invoke(Method theMethod, Object obj, List<String> args)
            throws IllegalAccessException, InvocationTargetException
    {
        log("invoking method: " + theMethod);
        log("            obj: " + obj);
        log("           args: " + args);

        Class<?>[] argTypes = theMethod.getParameterTypes();
        log("   looking for types: " + Arrays.toString(argTypes));
        if (argTypes.length != args.size())
            throw new IllegalArgumentException(theMethod.getName()
                    + "does not accept this many arguments");

        Object[] objArgs = new Object[args.size()];

        for (int i = 0; i < argTypes.length; ++i)
        {
            try
            {
                objArgs[i] = parseString(argTypes[i], args.get(i));
            }
            catch (Exception e)
            {
                throw new IllegalArgumentException(theMethod.getName()
                        + " expects a " + argTypes[i].getName()
                        + " for parameter " + i + ", but received "
                        + args.get(i));
            }
        }

        log("  passing these args: " + Arrays.toString(objArgs));

        return theMethod.invoke(obj, objArgs);
    }

    
    public void log(String s)
    {
        if (verbose)
            Harness.log("ReflectionFactory", s);
    }

    /**
     * Attempt to instantiate an object of class @a c that approximates the
     * string representation @a s.
     */
    @SuppressWarnings(value = "unchecked")
    private static Object parseString(Class type, String value)
            throws IllegalArgumentException
    {
        if (value == null)
            return null;
        if (Boolean.TYPE == type)
        {
            /* Boolean.parseBoolean(String) returns FALSE for anything other
             * than "true" (case insensitive). We want something a little more
             * robust than that.
             */
            if (value.equalsIgnoreCase("true"))
            {
                return Boolean.TRUE;
            }
            if (value.equalsIgnoreCase("false"))
            {
                return Boolean.FALSE;
            }
            throw new IllegalArgumentException("value \"" + value
                    + "\" could not" + " be parsed as a boolean value");
        }
        if (Byte.TYPE == type)
            return Byte.parseByte(value);
        if (Double.TYPE == type)
            return Double.parseDouble(value);
        if (Float.TYPE == type)
            return Float.parseFloat(value);
        if (Integer.TYPE == type)
            return Integer.parseInt(value);
        if (Long.TYPE == type)
            return Long.parseLong(value);
        if (Short.TYPE == type)
            return Short.parseShort(value);
        if (String.class == type)
            return value;
        if (SimTime.class.isAssignableFrom(type))
            return new SimTime(value);
        if (type.isEnum())
            /*
             * The next line will generate a compiler warning. I don't know how
             * to fix it.
             */
            return Enum.valueOf(type, value);
        throw new IllegalArgumentException("value \"" + value + "\" could not"
                + " be interpretted as " + type);
    }
}
